/*
	Copyright (C) 2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "tk_qt4.h"



jigsawCanvas :: jigsawCanvas ()
	:
	clickPieceID		( -1 ),
	currentTable		( 0 ),
	jigsaw			()
{
}

jigsawCanvas :: ~jigsawCanvas ()
{
	saveJigsaw ();
	mtjigsaw_destroy ( jigsaw );
	jigsaw = NULL;
}

void jigsawCanvas :: paintEvent (
	QPaintEvent	* event
	)
{
	mtImage		* image;
	int		px,
			py,
			pw,
			ph;
	QPainter	p ( this );



	px = event->rect ().x ();
	py = event->rect ().y ();
	pw = event->rect ().width ();
	ph = event->rect ().height ();

	if ( pw < 1 || ph < 1 )
	{
		return;
	}

	image = mtpixel_image_new_rgb ( pw, ph );
	if ( ! image )
	{
		return;
	}

	mtjigsaw_render ( jigsaw, image, px, py );

	QImage * im = new QImage (
		(const uchar *)image->channel[MTPIXEL_CHANNEL_IMAGE],
		pw, ph, pw * 3, QImage::Format_RGB888 );

	p.drawImage ( QPoint ( px, py ), im[0] );

	delete im;			// Delete before mtpixel image destroy

	mtpixel_image_destroy ( image );
}

void jigsawCanvas :: resizeEvent (
	QResizeEvent	* event
	)
{
	mtjigsaw_set_image ( jigsaw, NULL, width (), height () );
}

void jigsawCanvas :: mousePressEvent ( QMouseEvent * event )
{
	if ( event->button () == Qt::LeftButton )
	{
		int		ux,
				uy,
				uw,
				uh,
				piece_ID,
				x = event->x (),
				y = event->y ()
				;


		piece_ID = mtjigsaw_grab_piece ( jigsaw, x, y, &ux, &uy, &uw,
			&uh );

		if ( piece_ID < 0 )
		{
			// No piece has been clicked
			return;
		}

		update ( ux, uy, uw, uh );

		clickPieceID = piece_ID;
		clickX = x;
		clickY = y;
	}
}

void jigsawCanvas :: mouseMoveEvent ( QMouseEvent * event )
{
	if ( clickPieceID > -1 && event->buttons () & Qt::LeftButton )
	{
		int		x = event->x (),
				y = event->y (),
				dx,
				dy,
				ux,
				uy,
				uw,
				uh
				;


		dx = x - clickX;
		dy = y - clickY;

		if ( ! mtjigsaw_drag_piece ( jigsaw, clickPieceID, dx, dy, &ux,
			&uy, &uw, &uh )
			)
		{
			update ( ux, uy, uw, uh );
		}

		clickX = x;
		clickY = y;
	}
}

void jigsawCanvas :: mouseReleaseEvent ( QMouseEvent * event )
{
	if ( clickPieceID > -1 && event->button () == Qt::LeftButton )
	{
		int		ux,
				uy,
				uw,
				uh
				;


		if ( mtjigsaw_drop_piece ( jigsaw, clickPieceID, &ux, &uy, &uw,
			&uh ) == -1 )
		{
			update ( ux, uy, uw, uh );
		}

		clickPieceID = -1;
		saveJigsaw ();
	}
}

void jigsawCanvas :: saveJigsaw ()
{
	char		txt[2048];


	if ( ! jigsaw || ! currentTable )
	{
		return;
	}

	form_table_filename ( txt, 2048, currentTable );

	if ( mtjigsaw_save ( jigsaw, txt ) )
	{
		QMessageBox :: critical ( this, tr ( "Error" ),
			tr ( "Unable to save current Jigsaw." ) );
	}
}

void jigsawCanvas :: setJigsawTable (
	int	table
	)
{
	char		txt[2048];
	mtJigsaw	* new_jigsaw;


	if ( table < 1 || table > 10 )
	{
		table = 1;
	}

	saveJigsaw ();

	form_table_filename ( txt, 2048, table );
	new_jigsaw = mtjigsaw_load ( txt );

	if ( ! new_jigsaw )
	{
		new_jigsaw = mtjigsaw_load ( NULL );
		// Start with an empty jigsaw
	}

	prefs_set_int ( PREFS_MAIN_JIGSAW_TABLE, table );

	mtjigsaw_destroy ( jigsaw );
	jigsaw = new_jigsaw;
	currentTable = table;

	if ( ! jigsaw->nodes_hori )
	{
		setJigsawMesh ( prefs_get_int ( PREFS_MAIN_NODES_HORI ),
			prefs_get_int ( PREFS_MAIN_NODES_VERT ) );
	}

	mtjigsaw_set_image ( jigsaw, NULL, width (), height () );

	update ();
}

void jigsawCanvas :: setJigsawMesh (
	int	hori,
	int	vert
	)
{
	if (	hori < MTJIGSAW_NODES_MIN ||
		hori > MTJIGSAW_NODES_MAX ||
		vert < MTJIGSAW_NODES_MIN ||
		vert > MTJIGSAW_NODES_MAX
		)
	{
		hori = 10;
		vert = 10;
		prefs_set_int ( PREFS_MAIN_NODES_HORI, hori );
		prefs_set_int ( PREFS_MAIN_NODES_VERT, vert );
	}

	mtjigsaw_set_mesh ( jigsaw, hori, vert );

	update ();
}

void jigsawCanvas :: joinJigsaw ()
{
	mtjigsaw_join ( jigsaw );
	update ();
}

void jigsawCanvas :: scatterJigsaw ()
{
	mtjigsaw_scatter ( jigsaw );
	update ();
}

void jigsawCanvas :: setJigsawImage (
	QString		filename
	)
{
	if ( ! filename.isEmpty () )
	{
		if ( mtjigsaw_set_image ( jigsaw, filename.toUtf8 ().data (),
			width (), height () )
			)
		{
			QMessageBox :: critical ( this, tr ( "Error" ),
				tr ( "Unable to load image file." ) );
		}
		else
		{
			update ();
		}
	}
}

