/*
	Copyright (C) 2009-2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "private.h"



struct _mtPrefs
{
	mtTree		* tree;		// key = name (char *)
					// data = (mtPrefItemValue *)
	int		block_callbacks; // If set don't use callbacks
};



typedef enum		// Internal representation
{
	VTYPE_INT	= 1 << 0,
	VTYPE_DOUBLE	= 1 << 1,
	VTYPE_STR	= 1 << 2,

	VTYPE_ALL	= VTYPE_INT | VTYPE_DOUBLE | VTYPE_STR
} VType;



static void pitem_destroy (
	mtPrefItem	* item
	)
{
	free ( (void *)item->key );
	free ( (void *)item->def );
	free ( (void *)item->description );
	free ( (void *)item->opt );
	free ( item );
}



#define CHECK_DUP_STR( XXX ) if ( item->XXX ) \
	{ \
		new_item->XXX = strdup ( item->XXX ); \
		if ( ! new_item->XXX ) \
		{ \
			goto error; \
		} \
	}



static mtPrefItem * pitem_duplicate (
	mtPrefItem	* item
	)
{
	mtPrefItem	* new_item;


	new_item = malloc ( sizeof ( mtPrefItem ) );
	if ( ! new_item )
	{
		return NULL;
	}

	memcpy ( new_item, item, sizeof ( mtPrefItem ) );

	new_item->key = NULL;
	new_item->def = NULL;
	new_item->description = NULL;
	new_item->opt = NULL;

	CHECK_DUP_STR ( key )
	CHECK_DUP_STR ( def )
	CHECK_DUP_STR ( description )
	CHECK_DUP_STR ( opt )

	return new_item;

error:
	pitem_destroy ( new_item );

	return NULL;
}

static int mtkit_prefs_key_compare (
	void		* k1,
	void		* k2
	)
{
	return strcmp ( (char *)k1, (char *)k2 );
}

static void mtkit_prefs_node_destroy (
	mtTreeNode	* node
	)
{
	mtPrefItemValue	* item = (mtPrefItemValue *) node->data;


	// We don't free the key as it lives in the pref_item

	free ( item->value );
	pitem_destroy ( item->pref_item );
	free ( item );
}

static mtPrefItemValue * mtkit_prefs_get_value (
	mtPrefs		* prefs,
	const char	* key,
	VType		vtype
	)
{
	mtPrefItemValue	* piv;
	mtTreeNode	* node;


	if ( ! prefs || ! key )
	{
		return NULL;
	}

	node = mtkit_tree_node_find ( prefs->tree, (void *)key );
	if ( ! node )
	{
		return NULL;
	}

	piv = (mtPrefItemValue *)node->data;

	switch ( vtype )
	{
	case VTYPE_INT:
		if (	piv->pref_item->type == MTKIT_PREF_TYPE_INT	||
			piv->pref_item->type == MTKIT_PREF_TYPE_BOOL	||
			piv->pref_item->type == MTKIT_PREF_TYPE_RGB	||
			piv->pref_item->type == MTKIT_PREF_TYPE_OPTION
			)
		{
			break;
		}

		return NULL;		// Incompatible types

	case VTYPE_DOUBLE:
		if ( piv->pref_item->type == MTKIT_PREF_TYPE_DOUBLE )
		{
			break;
		}

		return NULL;		// Incompatible types

	case VTYPE_STR:
		if (	piv->pref_item->type == MTKIT_PREF_TYPE_STR	||
			piv->pref_item->type == MTKIT_PREF_TYPE_STR_MULTI ||
			piv->pref_item->type == MTKIT_PREF_TYPE_FILE	||
			piv->pref_item->type == MTKIT_PREF_TYPE_DIR
			)
		{
			break;
		}

		return NULL;		// Incompatible types

	case VTYPE_ALL:
		break;

	default:
		return NULL;		// Invalid type passed
	}

	return piv;
}

static int mtkit_prefs_set_value (
	mtPrefs		* prefs,
	const char	* key,
	const char	* value,
	VType		vtype,
	int		cb
	)
{
	mtPrefItemValue * piv;
	char		* v;


	if ( ! prefs || ! key )
	{
		return 1;
	}

	// Check this key exists and is of a compatible type
	piv = mtkit_prefs_get_value ( prefs, key, vtype );
	if ( ! piv )
	{
		return 2;
	}

	if ( ! value )
	{
		value = piv->pref_item->def;
	}

	if ( ! value )
	{
		value = "";
	}

	if (	vtype == VTYPE_STR &&
		piv->pref_item->opt
		)
	{
		int		max_chars;


		if ( mtkit_strtoi ( piv->pref_item->opt, &max_chars, NULL, 0 )
			|| max_chars < 1 )
		{
			return 2;
		}

		v = strdup ( value );

		if (	v &&
			strlen ( v ) > max_chars
			)
		{
			v[max_chars] = 0;
		}
	}
	else
	{
		v = strdup ( value );
	}

	if ( ! v )
	{
		return 3;
	}

	free ( piv->value );		// Lose old value
	piv->value = v;

	if (	cb				&&
		! prefs->block_callbacks	&&
		piv->pref_item->callback
		)
	{
		piv->pref_item->callback ( piv, piv->pref_item->callback_data,
			piv->pref_item->callback_ptr );
	}

	return 0;
}

mtPrefs * mtkit_prefs_new (
	mtPrefItem	* table
	)
{
	mtPrefs		* prefs;


	prefs = calloc ( sizeof ( mtPrefs ), 1 );
	if ( ! prefs )
	{
		return NULL;
	}

	prefs->tree = mtkit_tree_new ( mtkit_prefs_key_compare,
		mtkit_prefs_node_destroy );

	if ( ! prefs->tree )
	{
		goto error;
	}

	if ( mtkit_prefs_add ( prefs, table, 0 ) )
	{
		goto error;
	}

	return prefs;

error:
	mtkit_prefs_destroy ( prefs );

	return NULL;
}

int mtkit_prefs_add (
	mtPrefs		* prefs,
	mtPrefItem	* table,
	const char	* prefix
	)
{
	mtPrefItem	* item;
	mtPrefItemValue	* data;
	int		lenprefix = 0,
			lenkey;
	char		* new_key;


	if ( ! prefs || ! table )
	{
		return 1;
	}

	if ( prefix && prefix[0] )
	{
		lenprefix = strlen ( prefix );
	}

	// Populate the tree from the table
	for ( item = table; item->key; item ++ )
	{
		data = calloc ( sizeof ( mtPrefItemValue ), 1 );
		if ( ! data )
		{
			return 1;
		}

		if ( ! item->def )
		{
			data->value = strdup ( "" );
		}
		else
		{
			data->value = strdup ( item->def );
		}

		if ( ! data->value )
		{
			free ( data );

			return 1;
		}

		data->pref_item = pitem_duplicate ( item );
		if ( ! data->pref_item )
		{
			free ( data->value );
			free ( data );

			return 1;
		}

		if ( lenprefix )
		{
			lenkey = strlen ( data->pref_item->key );
			new_key = calloc ( lenkey + lenprefix + 1, 1 );

			if ( ! new_key )
			{
				pitem_destroy ( data->pref_item );
				free ( data->value );
				free ( data );

				return 1;
			}

			strcpy ( new_key, prefix );
			strcat ( new_key, data->pref_item->key );
			free ( (void *)data->pref_item->key );
			data->pref_item->key = new_key;
		}

		if ( ! mtkit_tree_node_add ( prefs->tree,
			(void *)data->pref_item->key, data ) )
		{
			pitem_destroy ( data->pref_item );
			free ( data->value );
			free ( data );

			return 1;
		}
	}

	return 0;
}

int mtkit_prefs_load (
	mtPrefs		* prefs,
	char		* filename
	)
{
	int		errors = 0;
	mtUtreeNode	* root,
			* node;


	if ( ! prefs || ! filename )
	{
		return 1;
	}

	root = mtkit_utree_load_file ( NULL, filename, &errors, 0 );
	if ( ! root )
	{
		return errors;
	}

	for ( node = root->child; node; node = node->next )
	{
		if (	node->type == MTKIT_UTREE_NODE_TYPE_ELEMENT	&&
			node->child					&&
			node->child->type == MTKIT_UTREE_NODE_TYPE_TEXT
			)
		{
			mtkit_prefs_set_value ( prefs, node->text,
				node->child->text, VTYPE_ALL, 0 );
		}
	}

	mtkit_utree_destroy_node ( root );

	return 0;
}

static int mtkit_prefs_save_recurse (
	mtUtreeNode	* root,
	mtTreeNode	* node
	)
{
	mtPrefItemValue	* data;


	if ( ! node )
	{
		return 0;
	}

	if ( mtkit_prefs_save_recurse ( root, node->left ) )
	{
		return 1;
	}

	data = (mtPrefItemValue	*)node->data;

	if (	( data->value[0] != 0 && ! data->pref_item->def ) ||
		( data->pref_item->def && strcmp ( data->value,
			data->pref_item->def ) )
		)
	{
		// Only save something if it deviates from the default

		mtUtreeNode	* key,
				* val;


		key = mtkit_utree_new_element ( root, (char *)node->key );
		if ( ! key )
		{
			return 1;
		}

		val = mtkit_utree_new_text ( key, data->value );
		if ( ! val )
		{
			return 1;
		}
	}

	if ( mtkit_prefs_save_recurse ( root, node->right ) )
	{
		return 1;
	}

	return 0;
}

int mtkit_prefs_save (
	mtPrefs		* prefs,
	char		* filename
	)
{
	int		error = 0;
	mtUtreeNode	* root;


	root = mtkit_utree_new_root ();
	if ( ! root )
	{
		return 0;
	}

	// Populate the utree
	error = mtkit_prefs_save_recurse ( root, prefs->tree->root );

	if ( ! error )
	{
		error = mtkit_utree_save_file ( root, filename,
			MTKIT_UTREE_OUTPUT_DEFAULT, 0 );
	}

	mtkit_utree_destroy_node ( root );

	return error;
}

int mtkit_prefs_destroy (
	mtPrefs		* prefs
	)
{
	if ( ! prefs )
	{
		return 1;
	}

	mtkit_tree_destroy ( prefs->tree );

	free ( prefs );

	return 0;
}

int mtkit_prefs_block_callback (
	mtPrefs		* prefs
	)
{
	if ( ! prefs )
	{
		return 1;
	}

	prefs->block_callbacks = 1;

	return 0;
}

int mtkit_prefs_unblock_callback (
	mtPrefs		* prefs
	)
{
	if ( ! prefs )
	{
		return 1;
	}

	prefs->block_callbacks = 0;

	return 0;
}

int mtkit_prefs_get_int (
	mtPrefs		* prefs,
	const char	* key,
	int		* value
	)
{
	mtPrefItemValue	* piv;


	if ( ! prefs || ! key || ! value )
	{
		return 1;
	}

	piv = mtkit_prefs_get_value ( prefs, key, VTYPE_INT );
	if ( ! piv )
	{
		return 2;		// Not in tree/table or wrong type
	}

	return mtkit_strtoi ( piv->value, value, NULL, 0 );
}

int mtkit_prefs_get_double (
	mtPrefs		* prefs,
	const char	* key,
	double		* value
	)
{
	mtPrefItemValue	* piv;


	if ( ! prefs || ! key || ! value )
	{
		return 1;
	}

	piv = mtkit_prefs_get_value ( prefs, key, VTYPE_DOUBLE );
	if ( ! piv )
	{
		return 2;		// Not in tree/table or wrong type
	}

	return mtkit_strtod ( piv->value, value, NULL, 0 );
}

int mtkit_prefs_get_str (
	mtPrefs		* prefs,
	const char	* key,
	char		** value
	)
{
	mtPrefItemValue	* piv;


	if ( ! prefs || ! key || ! value )
	{
		return 1;
	}

	piv = mtkit_prefs_get_value ( prefs, key, VTYPE_STR );
	if ( ! piv )
	{
		return 2;		// Not in tree/table or wrong type
	}

	value[0] = piv->value;

	return 0;
}

int mtkit_prefs_set_default (
	mtPrefs		* prefs,
	const char	* key
	)
{
	if ( ! prefs || ! key )
	{
		return 1;
	}

	return mtkit_prefs_set_value ( prefs, key, NULL, VTYPE_ALL, 1 );
}

int mtkit_prefs_set_int (
	mtPrefs		* prefs,
	const char	* key,
	int		value
	)
{
	char		str[256];


	if ( ! prefs || ! key )
	{
		return 1;
	}

	sprintf ( str, "%i", value );

	return mtkit_prefs_set_value ( prefs, key, str, VTYPE_INT, 1 );
}

int mtkit_prefs_set_double (
	mtPrefs		* prefs,
	const char	* key,
	double		value
	)
{
	char		str[256];


	if ( ! prefs || ! key )
	{
		return 1;
	}

	sprintf ( str, "%.15lg", value );

	return mtkit_prefs_set_value ( prefs, key, str, VTYPE_DOUBLE, 1 );
}

int mtkit_prefs_set_str (
	mtPrefs		* prefs,
	const char	* key,
	const char	* value
	)
{
	if ( ! prefs || ! key )
	{
		return 1;
	}

	return mtkit_prefs_set_value ( prefs, key, value, VTYPE_STR, 1 );
}

char * mtkit_prefs_type_text (
	mtPrefType	type
	)
{
	static char * ptypes[MTKIT_PREF_TYPE_TOTAL] = {
			"?",
			_("integer"),
			_("boolean"),
			_("RGB"),
			_("option"),
			_("decimal"),
			_("string"),
			_("string multi-line"),
			_("filename"),
			_("directory")
			};

	if (	type <= MTKIT_PREF_TYPE_NONE ||
		type >= MTKIT_PREF_TYPE_TOTAL
		)
	{
		return NULL;
	}

	return ptypes[type];
}

void mtkit_prefs_get_str_val (
	mtPrefItemValue	* piv,
	const char	* value,
	char		* buf,
	int		buf_size
	)
{
	const char	* txt = NULL;
	int		num;


	buf[0] = 0;		// Default to empty string for sanity
	if ( ! value )
	{
		return;
	}

	switch ( piv->pref_item->type )
	{
	case MTKIT_PREF_TYPE_BOOL:
		if ( value[0] == '0' )
		{
			txt = _("FALSE");
		}
		else
		{
			txt = _("TRUE");
		}

		break;

	case MTKIT_PREF_TYPE_RGB:
		if ( mtkit_strtoi ( value, &num, NULL, 0 ) )
		{
			txt = "?";
		}
		else
		{
			snprintf ( buf, buf_size, "( %i , %i , %i )",
				( (num >> 16) & 0xFF ),
				( (num >> 8) & 0xFF ),
				( num & 0xFF )
				);

			// Only needed for small buf_size
			buf[buf_size - 1] = 0;
		}
		break;

	case MTKIT_PREF_TYPE_OPTION:
		if ( mtkit_strtoi ( value, &num, NULL, 0 ) )
		{
			txt = "?";
		}
		else
		{
			char		* txt_option;


			txt_option = mtkit_strtok ( piv->pref_item->opt, "\t",
				num );

			if ( txt_option )
			{
				snprintf ( buf, buf_size, "( %i ) = %s", num,
					txt_option );
				free ( txt_option );
			}
			else
			{
				snprintf ( buf, buf_size, "( %i ) = ?", num );
			}

			// Only needed for small buf_size
			buf[buf_size - 1] = 0;
		}
		break;

	default:
		txt = value;
		break;
	}

	if ( txt )
	{
		mtkit_strncpy0 ( buf, txt, buf_size );
	}
}



#define BULK_LOOP_GET( TABLE, FUNCTION ) \
	if ( TABLE ) \
	{ \
		for ( ; TABLE->name; TABLE ++ ) \
		{ \
			if ( TABLE->var ) \
			{ \
				FUNCTION ( prefs, TABLE->name, TABLE->var ); \
			} \
		}\
	}



int mtkit_prefs_bulk_get (
	mtPrefs		* prefs,
	mtBulkInt	* table_i,
	mtBulkDouble	* table_d,
	mtBulkStr	* table_s
	)
{
	BULK_LOOP_GET ( table_i, mtkit_prefs_get_int )
	BULK_LOOP_GET ( table_d, mtkit_prefs_get_double )

	if ( table_s )
	{
		char		* new;


		for ( ; table_s->name; table_s ++ )
		{
			if ( table_s->var )
			{
				if ( ! mtkit_prefs_get_str ( prefs,
					table_s->name, &new ) )
				{
					if ( mtkit_strfreeswap ( table_s->var,
						new ) )
					{
						return 1;
					}
				}
			}
		}
	}

	return 0;
}



#define BULK_LOOP_SET( TABLE, FUNCTION ) \
	if ( TABLE ) \
	{ \
		for ( ; TABLE->name; TABLE ++ ) \
		{ \
			if ( TABLE->var ) \
			{ \
				if ( FUNCTION ( prefs, TABLE->name, \
					TABLE->var[0] ) ) \
				{ \
					return 1; \
				} \
			} \
		} \
	}



int mtkit_prefs_bulk_set (
	mtPrefs		* prefs,
	mtBulkInt	* table_i,
	mtBulkDouble	* table_d,
	mtBulkStr	* table_s
	)
{
	BULK_LOOP_SET ( table_i, mtkit_prefs_set_int )
	BULK_LOOP_SET ( table_d, mtkit_prefs_set_double )
	BULK_LOOP_SET ( table_s, mtkit_prefs_set_str )

	return 0;
}

int mtkit_prefs_value_mirror (
	mtPrefs		* dest,
	mtPrefs		* src,
	mtPrefItem	* table
	)
{
	int		i,
			ival;
	char		* sval;
	double		dval;


	if ( ! dest || ! src || ! table )
	{
		goto fail;
	}

	for ( i = 0; table[i].key; i++ )
	{
		switch ( table[i].type )
		{
		case MTKIT_PREF_TYPE_INT:
		case MTKIT_PREF_TYPE_BOOL:
		case MTKIT_PREF_TYPE_RGB:
		case MTKIT_PREF_TYPE_OPTION:
			if ( mtkit_prefs_get_int ( src, table[i].key, &ival ) ||
				mtkit_prefs_set_int ( dest, table[i].key, ival )
				)
			{
				goto fail;
			}

			break;

		case MTKIT_PREF_TYPE_DOUBLE:
			if ( mtkit_prefs_get_double ( src, table[i].key, &dval )
				||
				mtkit_prefs_set_double ( dest, table[i].key,
					dval )
				)
			{
				goto fail;
			}

			break;

		case MTKIT_PREF_TYPE_STR:
		case MTKIT_PREF_TYPE_STR_MULTI:
		case MTKIT_PREF_TYPE_FILE:
		case MTKIT_PREF_TYPE_DIR:
			if ( mtkit_prefs_get_str ( src, table[i].key, &sval ) ||
				mtkit_prefs_set_str ( dest, table[i].key, sval )
				)
			{
				goto fail;
			}

			break;

		default:
			goto fail;
		}
	}

	return 0;

fail:
	return 1;
}

int mtkit_prefs_value_copy (
	mtPrefs		* dest,
	mtPrefs		* src,
	mtPrefTrans	* table
	)
{
	int		i;
	mtPrefItemValue	* item_src,
			* item_dest;
	mtTreeNode	* tnode;


	if ( ! dest || ! src || ! table )
	{
		goto fail;
	}

	for ( i = 0; table[i].src && table[i].dest; i++ )
	{
		// Get source data
		tnode = mtkit_tree_node_find ( src->tree,
			(void *)table[i].src );

		if ( ! tnode || ! tnode->data )
		{
			goto fail;
		}

		item_src = tnode->data;

		// Get destination data
		tnode = mtkit_tree_node_find ( dest->tree,
			(void *)table[i].dest );

		if ( ! tnode || ! tnode->data )
		{
			goto fail;
		}

		item_dest = tnode->data;

		// Type validation
		if ( item_src->pref_item->type != item_dest->pref_item->type )
		{
			goto fail;
		}

		// Duplicate value
		if ( mtkit_strfreeswap ( &item_dest->value, item_src->value ) )
		{
			goto fail;
		}
	}

	return 0;

fail:
	return 1;
}

mtTree * mtkit_prefs_get_tree (
	mtPrefs		* prefs
	)
{
	if ( ! prefs )
	{
		return NULL;
	}

	return prefs->tree;
}
