/*
	Copyright (C) 2013 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#include "tk_qt4.h"
#include "icon_xpm.xpm"



mtTableAnalysis :: mtTableAnalysis (
	CedSheet	* sheetData,
	const char	* pathAnalysed,
	MainWindow	* mainwin,
	QVBoxLayout	* layout
	)
	:
	sortColumn	( RAFT_COL_BYTES ),
	sheetRowTotal	( 0 ),
	sortDirection	( Qt::DescendingOrder ),
	path		( NULL ),
	sheet		( sheetData ),
	mainWindow	( mainwin )
{
	unsigned int	rowtot;
	int		i,
			width;
	QTableWidgetItem * twItem;
	QStringList	columnLabels;
	QLineEdit	* pathEdit;


	if ( pathAnalysed )
	{
		path = strdup ( pathAnalysed );
	}

	if ( ! path )
	{
		fprintf ( stderr, "Error (mtTableAnalysis): path is NULL\n" );
	}

	ced_sheet_get_geometry ( sheet, &rowtot, NULL );
	sheetRowTotal = rowtot;

	setSelectionMode ( QAbstractItemView::SingleSelection );
	setSelectionBehavior ( QAbstractItemView::SelectRows );
	setEditTriggers ( QAbstractItemView::NoEditTriggers );
	setColumnCount ( RAFT_COL_TOTAL - 1 );

	columnLabels
		<< tr ( "Name" )
		<< tr ( "Files" )
		<< tr ( "%" )
		<< tr ( "Bytes" )
		<< tr ( "MB" )
		<< tr ( "%" )
		<< tr ( "Subdirs" )
		<< tr ( "Other" )
		;
	setHorizontalHeaderLabels ( columnLabels );

	twItem = horizontalHeaderItem ( 0 );
	if ( twItem )
	{
		twItem->setTextAlignment ( Qt::AlignLeft | Qt::AlignVCenter );
	}

	for ( i = 1; i < (RAFT_COL_TOTAL - 1); i++ )
	{
		twItem = horizontalHeaderItem ( i );
		if ( twItem )
		{
			twItem->setTextAlignment ( Qt::AlignRight |
				Qt::AlignVCenter );
		}
	}

	horizontalHeader ()->setSortIndicatorShown ( true );
	sortTable ( sortColumn, sortDirection );

	verticalHeader ()->hide ();
	verticalHeader ()->setDefaultSectionSize (
		verticalHeader ()->fontMetrics ().height () + 4 );

	setShowGrid ( false );

	connect ( horizontalHeader (), SIGNAL ( sectionClicked ( int ) ),
		this, SLOT ( tableHeaderClick ( int ) ) );

	connect ( this, SIGNAL ( cellActivated ( int, int ) ),
		this, SLOT ( tableCellActivated ( int, int ) ) );

	horizontalHeader ()->resizeSections ( QHeaderView::ResizeToContents );

	// Add some padding to columns
	for ( i = 0; i < (RAFT_COL_TOTAL - 1); i++ )
	{
		width = horizontalHeader ()->sectionSize ( i );
		horizontalHeader ()->resizeSection ( i, width + 8 );
	}

	pathEdit = new QLineEdit ( mtQEX :: qstringFromC (
		(char *)pathAnalysed ) );
	pathEdit->setReadOnly ( true );

	layout->addWidget ( pathEdit );
	layout->addWidget ( this );
}

mtTableAnalysis :: ~mtTableAnalysis ()
{
	free ( path );
	path = NULL;

	ced_sheet_destroy ( sheet );
	sheet = NULL;
}

void mtTableAnalysis :: copyToClipboard ()
{
	char		* txt;


	txt = raft_get_clipboard ( sheet );
	if ( txt )
	{
		QApplication :: clipboard ()->setText ( txt );

		free ( txt );
	}
}

void mtTableAnalysis :: sortTable (
	int		column,
	Qt::SortOrder	direction
	)
{
	unsigned int	cols[] = { 1, 0 };
	int		r,
			c;
	char		* csp;
	CedCell		* cell;
	QTableWidgetItem * twItem;


	if ( column < 1 )
	{
		column = -column;

		if ( column == sortColumn )
		{
			// Toggle the current sort direction
			if ( sortDirection == Qt::DescendingOrder )
			{
				direction = Qt::AscendingOrder;
			}
			else
			{
				direction = Qt::DescendingOrder;
			}
		}
		else
		{
			direction = Qt::DescendingOrder;
		}
	}

	horizontalHeader ()->setSortIndicator ( (column - 1), direction );

	sortColumn = column;
	sortDirection = direction;

	r = 0;
	if ( sortDirection == Qt::DescendingOrder )
	{
		r = CED_SORT_MODE_DESCENDING;
	}

	cols[0] = sortColumn;
	ced_sheet_sort_rows ( sheet, 1, 0, cols, r, NULL );

	setCurrentItem ( NULL );	// Stops double selection of 0,0
	clearContents ();
	setRowCount ( sheetRowTotal );

	for ( r = 1; r <= sheetRowTotal; r++ )
	{
		for ( c = 1; c < RAFT_COL_TOTAL; c++ )
		{
			cell = ced_sheet_get_cell ( sheet, r, c );
			if ( ! cell )
			{
				// Should never happen
				continue;
			}

			csp = ced_cell_create_output ( cell, NULL );
			if ( ! csp )
			{
				continue;
			}

			twItem = new QTableWidgetItem;
			twItem->setText ( mtQEX :: qstringFromC ( csp ) );
			free ( csp );

			if ( c > 1 )
			{
				twItem->setTextAlignment ( Qt::AlignRight |
					Qt::AlignVCenter );
			}

			setItem ( r - 1, c - 1, twItem );
		}
	}

	setCurrentCell ( 0, 0 );
}

void mtTableAnalysis :: tableHeaderClick (
	int		index
	)
{
	sortTable ( -(index + 1), Qt::DescendingOrder );
}

void mtTableAnalysis :: tableCellActivated (
	int		row,
	int		column
	)
{
	char		buf [ PATH_LEN ];


	if ( 0 == raft_path_merge ( buf, path, sheet, row + 1 ) )
	{
		mainWindow->doAnalysis ( buf );
	}
}

MainWindow :: MainWindow ()
	:
	progressBarValue	( 0 ),
	table			()
{
	QWidget		* widget;
	QVBoxLayout	* layout;
	QHBoxLayout	* topRow;
	QPushButton	* buttonQuit;
	QShortcut	* shortcut;


	setWindowTitle ( VERSION );
	this->setWindowIcon ( QPixmap ( icon_xpm ) );

	widget = new QWidget;
	setCentralWidget ( widget );

	shortcut = new QShortcut ( QKeySequence ( Qt::CTRL + Qt::Key_PageUp ),
		this );
	connect ( shortcut, SIGNAL ( activated () ), this,
		SLOT ( pressTabNext () ) );

	shortcut = new QShortcut ( QKeySequence ( Qt::CTRL +
		Qt::Key_PageDown ), this );
	connect ( shortcut, SIGNAL ( activated () ), this,
		SLOT ( pressTabPrevious () ) );

	buttonQuit = new QPushButton ( tr ( "Quit" ) );
	buttonQuit->setShortcut ( Qt :: CTRL + Qt :: Key_Q );
	connect ( buttonQuit, SIGNAL ( clicked () ), this,
		SLOT ( pressButtonQuit () ) );

	progressBar = new QProgressBar;
	progressBar->setMaximum ( PROGRESS_MAX );
	setProgress ( 0 );

	buttonCopy = new QPushButton ( tr ( "Copy To Clipboard" ) );
	buttonCopy->setShortcut ( Qt :: CTRL + Qt :: Key_C );
	connect ( buttonCopy, SIGNAL ( clicked () ), this,
		SLOT ( pressButtonCopy () ) );

	buttonCopy->setEnabled ( false );

	topRow = new QHBoxLayout ();
	topRow->addWidget ( buttonQuit );
	topRow->addWidget ( progressBar );
	topRow->addWidget ( buttonCopy );

	tabWidget = new QTabWidget;

	layout = new QVBoxLayout;
	layout->setMargin ( 5 );
	layout->addLayout ( topRow );
	layout->addWidget ( tabWidget );

	widget->setLayout ( layout );

	prefs_init ();
	prefs_load ();

	setMinimumSize ( 160, 160 );

	setGeometry ( prefs_get_int ( PREFS_WINDOW_X ),
		prefs_get_int ( PREFS_WINDOW_Y ),
		prefs_get_int ( PREFS_WINDOW_W ),
		prefs_get_int ( PREFS_WINDOW_H ) );

	show ();
}

MainWindow :: ~MainWindow ()
{
	prefs_set_int ( PREFS_WINDOW_X, geometry().x () );
	prefs_set_int ( PREFS_WINDOW_Y, geometry().y () );
	prefs_set_int ( PREFS_WINDOW_W, geometry().width () );
	prefs_set_int ( PREFS_WINDOW_H, geometry().height () );

	prefs_save ();
	prefs_close ();

	for ( int i = 0; i < TABLE_MAX; i++ )
	{
		if ( table[i] == NULL )
		{
			continue;
		}

		delete table[i];
		table[i] = NULL;
	}
}

static int raftScan (
	CedSheet	* sheet,
	int		row,		// Row calculated in sheet
	void		* user_data
	)
{
	MainWindow	* main = (MainWindow *)user_data;


	main->setProgress ( PROGRESS_NEXT );

	while ( QCoreApplication :: hasPendingEvents () )
	{
		QCoreApplication :: processEvents ();
	}

	if ( main->wasStopped () )
	{
		return 1;		// User wants to stop
	}

	return 0;			// Keep scanning
}

int MainWindow :: doAnalysis (
	const char	* path
	)
{
	char		pbuf [ PATH_LEN ],
			pbuf_short [ PATH_LEN_SHORT_BUF ];
	int		tabTot,
			tabCurrent,
			res = 0;
	CedSheet	* sheet;


	tabTot = tabWidget->count ();
	tabCurrent = tabWidget->currentIndex ();

	if (	tabTot < 0		||
		tabTot >= TABLE_MAX	||
		table[ tabTot ] != NULL
		)
	{
		// Should never happen, so silently fail

		return 1;
	}

	if ( raft_path_check ( path, pbuf ) )
	{
		return 1;
	}

	tabWidget->setEnabled ( false );
	stopFlag = 0;

	if ( raft_scan_sheet ( pbuf, &sheet, raftScan, this ) )
	{
		res = 1;		// Fail
	}
	else
	{
		QWidget		* widget;
		QVBoxLayout	* layout;


		snprintf ( pbuf_short, sizeof ( pbuf_short ),
			"%i", tabTot + 1 );

		widget = new QWidget;
		tabWidget->addTab ( widget,
			mtQEX :: qstringFromC ( pbuf_short ) );
		tabWidget->setCurrentWidget ( widget );

		layout = new QVBoxLayout;
		layout->setMargin ( 0 );
		widget->setLayout ( layout );

		table[ tabTot ] = new mtTableAnalysis ( sheet, pbuf, this,
			layout );

		buttonCopy->setEnabled ( true );
	}

	tabWidget->setEnabled ( true );
	setProgress ( 0 );

	if ( tabCurrent >= 0 && tabCurrent < TABLE_MAX )
	{
	/*	This hack is required because:

		After successfully creating a new tab, stop focus on the old tab
		moving to the QLineEdit widget.

		Keep the table selected when there is a failure resulting from
		trying to scan the <.> or <TOTAL> rows.
	*/

		table[ tabCurrent ]->setFocus ();
	}

	if ( res == 0 )
	{
		table[ tabTot ]->setFocus ();
	}

	return res;
}

void MainWindow :: pressButtonQuit ()
{
	if ( isBusy () )
	{
		// We are analysing so stop
		stopNow ();

		return;
	}

	close ();
}

void MainWindow :: pressButtonCopy ()
{
	int		tabNum;


	tabNum = tabWidget->currentIndex ();

	if (	tabNum >= 0		&&
		tabNum < TABLE_MAX	&&
		table[ tabNum ] != NULL
		)
	{
		table[ tabNum ]->copyToClipboard ();
	}
}

void MainWindow :: pressTabNext ()
{
	tabWidget->setCurrentIndex ( tabWidget->currentIndex () - 1 );
}

void MainWindow :: pressTabPrevious ()
{
	tabWidget->setCurrentIndex ( tabWidget->currentIndex () + 1 );
}

void MainWindow :: setProgress (
	int		val
	)
{
	if ( val < 0 || val > PROGRESS_MAX )
	{
		progressBarValue = ( progressBarValue % PROGRESS_MAX ) + 1;
	}
	else
	{
		progressBarValue = val;
	}

	progressBar->setValue ( progressBarValue );

	if ( val == 0 )
	{
		progressBar->setFormat ( tr ( "Status: Idle" ) );
	}
	else
	{
		progressBar->setFormat ( tr ( "Status: Working" ) );
	}
}

int MainWindow :: isBusy ()
{
	return progressBarValue;
}

void MainWindow :: stopNow ()
{
	stopFlag = 1;
}

int MainWindow :: wasStopped ()
{
	return stopFlag;
}

void MainWindow :: closeEvent (
	QCloseEvent	* event
	)
{
	if ( isBusy () )
	{
		// Program is busy so we can't stop yet, but signal a stop
		stopNow ();
		event->ignore ();
	}
	else
	{
		// Program is idle so accept closure
		event->accept ();
	}
}

int main (
	int		argc,
	char		** argv
	)
{
	const char	* scan_directory = raft_cline ( argc, argv );
	QApplication	app ( argc, argv );
	MainWindow	window;


	if ( scan_directory )
	{
		window.doAnalysis ( scan_directory );
	}

	return app.exec ();
}
